EMPTY_CALL: str = "Received empty remote procedure call data."
CALL_INTERCEPTED: str = "Received RPC: %s"
RPC_METHOD_CALL_START: str = "Calling the '%s' method with RPC ID #%s."
RPC_METHOD_CALL_END: str = "RPC ID #%s for method '%s' processed with result: %s."
RPC_NOTIFICATION_START: str = "Notification method '%s' received."
RPC_NOTIFICATION_END: str = "Notification method '%s' processed."
INVALID_JSON_RPC_VERSION: str = "Invalid JSON-RPC version! Expected '2.0', got '%s'."

# flakes8: noqa: E501
